
const RAW = require('raw-device');
const fs = require('fs');
const xml2js = require('xml2js');

/* NECD object uses RAW object as its prototype */
NECD.prototype = Object.create(RAW.prototype);
NECD.prototype.constructor = NECD;

NECD.prototype.addressDefaults = {
    name: 'NECDisplay',
	mode: 'tcp',
    id: 1,
	//tcp
	port: 7142,
	//serial
	baudRate: 9600,
	dataBits:8,
	parity: 'none',
	stopBits: 1
}

NECD.prototype.optionsDefaults = {
    disconnect: true,
	encoding: 'ASCII',
	writeDuration: 500,
	readDuration: 500,
    dataname: 'necd.xml',
	splitter: {
		 delimiter: '\r',
		 includeDelimiter: true
	}
}

/**
 * Constructor for NECD object
 * @constructor
 * @param {AddressObject} address
 * @param {OptionsObject} [options]
 */
function NECD(address, options={}) {
    RAW.call(this, address, options);
	this.id = this.address.id;
	this.encodedId = encodeID(this.id);
	this.data;
	let dataName = this.options.dataname;
    let parser = new xml2js.Parser();
    let dataXML = fs.readFileSync(__dirname + '/' + dataName);
    parser.parseString(dataXML, (err, json) => {
        this.data = json;
    });
}

/**
 * Encode command for device using its communication protocol
 * @param {string} cmd  - command to encode
 * @returns {CommandObject} cmdObj
 */
NECD.prototype.encode = function(cmd){
    if(cmd.startsWith('#')){
		let co = this.special(cmd);
        return co;
	}
    let res = /(\w+)(\?)?( ([\w-]+)(,(\w+)?)?)?/.exec(cmd);
	if(!res)
		return;
    let name = res[1];
	let cmdType = ''; //param, command

	if(this.data.dev.param.find(el => el['$'].name.toLowerCase() == name.toLowerCase()))
		cmdType = 'param';
	else if(this.data.dev.command.find(el => el['$'].name.toLowerCase() == name.toLowerCase()))
		cmdType = 'command';
	else 
		return;
	let mode = res[2];
	let value = res[4];	
	let cmdObj;
	if(cmdType == 'param'){
		if(mode == '?') //get
			cmdObj = this.paramGet(name);
		else //set
			cmdObj = this.paramSet(name, value);
	}
	else if(cmdType == 'command'){
		if(mode == '?'){// command read
			cmdObj = this.commandRead(name);
		} 
		else {//command write
			cmdObj = this.commandWrite(name, value)
		}
	}
	return cmdObj;
}

NECD.prototype.commandWrite = function(name, value){//poprawić i przetestować
	let cmdef = this.data.dev.command.find(item => item['$'].name.toLowerCase() == name.toLowerCase());
	if(!cmdef)
        return;
	if(!cmdef['$'].mode.includes('w'))
		return;
	let cmdObj = {
		dev: this.name,
		command: name
	}
	let dur = parseInt(cmdef['$'].writeDuration);
	cmdObj['duration'] = dur? dur: this.options.writeDuration;
	let code = cmdef['$'].code;
	let val = parseInt(value);
	if(isNaN(val)){
		if(!('item' in cmdef))
			return;
		let sel = cmdef.item.find(el => el['$'].name.toLowerCase() == value.toLowerCase());
		val = parseInt(sel['$'].value);
	}
	let valDigits = parseInt(cmdef['$'].valdigits); //! przemyśleć jeszcze valdigits
	let valhex = decimal2Hex(val, valDigits);
	let mlength = decimal2Hex(code.length + valDigits +2);
	let commandStr = `\x010${this.encodedId}0A${mlength}\x02${code}${valhex}\x03`;
	let bcc = calcBCC(commandStr);
	commandStr += `${String.fromCharCode(bcc)}\x0D`;
	cmdObj['encodedstr'] = commandStr;
	cmdObj['encoded'] = Buffer.from(commandStr, this.encoding);
	return cmdObj;
}

NECD.prototype.commandRead = function(name){
	let cmdef = this.data.dev.command.find(item => item['$'].name.toLowerCase() == name.toLowerCase());
	if(!cmdef)
        return;
	if(!cmdef['$'].mode.includes('r'))
	return;
	let cmdObj = {
		dev: this.name,
		command: name
	}
	let dur = parseInt(cmdef['$'].readDuration);
	cmdObj['duration'] = dur? dur: this.options.readDuration;
	let code = cmdef['$'].code;
	let mlength = decimal2Hex(code.length +2);
	let commandStr = `\x010${this.encodedId}0A${mlength}\x02${code}\x03`;
	let bcc = calcBCC(commandStr);
	commandStr += `${String.fromCharCode(bcc)}\x0D`;
	cmdObj['encodedstr'] = commandStr;
	cmdObj['encoded'] = Buffer.from(commandStr, this.encoding);
	return cmdObj;
}

NECD.prototype.paramGet = function(name){
	let cmdef = this.data.dev.param.find(el => el['$'].name.toLowerCase() == name.toLowerCase());
	if(!cmdef)
        return;
	let cmdObj = {
		dev: this.name,
		command: name
	}
	let dur = parseInt(cmdef['$'].readDuration);
	cmdObj['duration'] = dur? dur: this.options.readDuration;
	let code = cmdef['$'].code;
	let mlength = decimal2Hex(1 + code.length + 1);
	var commandStr = `\x010${this.encodedId}0C${mlength}\x02${code}\x03`;
	var bcc = calcBCC(commandStr);
	commandStr += `${String.fromCharCode(bcc)}\x0D`;
	cmdObj['encodedstr'] = commandStr;
	cmdObj['encoded'] = Buffer.from(commandStr, this.encoding);
	return cmdObj;
}

NECD.prototype.paramSet = function(name, value){
	let cmdef = this.data.dev.param.find(el => el['$'].name.toLowerCase() == name.toLowerCase());
	if(value === undefined)
		return;
	if(!cmdef)
        return;
	let cmdObj = {
		dev: this.name,
		command: name
	}
	let dur = parseInt(cmdef['$'].writeDuration);
	cmdObj['duration'] = dur? dur: this.options.writeDuration;
	let code = cmdef['$'].code;
	let mlength = decimal2Hex(1 + code.length + 4 + 1);
	let val = parseInt(value);
	if(isNaN(val)){
		if(! ('item' in cmdef))
			return;
		let sel = cmdef.item.find(el => el['$'].name.toLowerCase() == value.toLowerCase());
		if(!sel)
			return;
		val = parseInt(sel['$'].value);
	}
	var commandStr = `\x010${this.encodedId}0E${mlength}` + `\x02${code}${decimal2Hex(val,4)}\x03`;
	var bcc = calcBCC(commandStr);
	commandStr += `${String.fromCharCode(bcc)}\x0D`;
	cmdObj['encodedstr'] = commandStr;
	cmdObj['encoded'] = Buffer.from(commandStr, this.encoding);
	return cmdObj;
}
/**
 * Decode response from device to friendly form
 * @param {Buffer} data - a data from splitter
 * @fires NECD#responseFromDevice
 */
NECD.prototype.decode = function(data){
	let str = data.toString(this.options.encoding);
	let response = {
        dev: this.name,
        raw: data,
    }
	let result = /\x0100(?<id>.)(?<mtype>[A-F])(?<mlength>[A-Fa-f0-9]{2})\x02(?<message>[^\x03]+)\x03/.exec(data);//(?<bcc>[A-Fa-f0-9]{1})\x0D/;
    if(!result){
		response['value'] = "can't decode response";
		this.emitter.emit('responseFromDevice', response);
		return response;
	}
	let id = result.groups.id;
	if(id != this.encodedId) //not for me response. Can happen when RS232 chain is used
		return;
	response['id'] = decodeID(id);

	let mType = result.groups.mtype;
	let allvalues = {
			msgtype: mType
	};
	response['allvalues'] = allvalues;
	if((mType == 'D') || (mType == 'F')){//GET_PARAMETER_REPLY //SET_PARAMETER_REPLY
		let resm = /(?<result>[A-Fa-f0-9]{2})(?<param>[A-Fa-f0-9]{4})(?<type>[A-Fa-f0-9]{2})(?<max>[A-Fa-f0-9]{4})(?<value>[A-Fa-f0-9]{4})/.exec(result.groups.message);
		let code = resm.groups.param;
		response['req'] = code;
		allvalues['code'] = code;
		allvalues['value'] = resm.groups.value;
		allvalues['max'] = resm.groups.max;
		allvalues['type'] = resm.groups.type;
		let value = parseInt(resm.groups.value, 16);	
		response['value'] = value;

		let cmdef = this.data.dev.param.find(el => el['$'].code == code);
		if(!cmdef){
			this.emitter.emit('responseFromDevice', response);
			return response;
		}
		response['req'] = cmdef['$'].name;
		//some additional claculations for some params
		if(response.req == 'temperature')
			response['value'] = value/2;
		else if(response.req == 'colorTemp')
			response['value'] = 2600 + value*100;

		let items = cmdef.item;
		if(!items){
			this.emitter.emit('responseFromDevice', response);
			return response;
		}
		let itemValue = items.find(el => parseInt(el['$'].value) == value)
		response['value'] = itemValue['$'].name;
	}
	else if(mType == 'B'){//COMMAND_REPLY
		let msg = result.groups.message;
		let cmdef = this.data.dev.command.find(el =>{
			let re = new RegExp(el['$'].pattern);
			let res = re.exec(msg);
			if(res)
				return true;
		})
		response['req'] = cmdef['$'].req? cmdef['$'].req: cmdef['$'].name;
		let re = new RegExp(cmdef['$'].pattern);
		let res = re.exec(msg);
		allvalues['value'] = res[1];

		let ret = cmdef['$'].return;
		if(ret == 'number'){ //response value is number
			let val = parseInt(res[1], 16);
			if(cmdef.item){
				let sel = cmdef.item.find(el => parseInt(el['$'].value) == val);
				if(sel)
					val = sel['$'].name;
			}
			response['value'] = val;
		}
		else if(ret == 'string'){
			let s = asciiHex2str(res[1]);
			response['value'] = s.trim();
		}
	}
	this.emitter.emit('responseFromDevice', response);
	return response;
}

function calcBCC(ascii){
	let bcc = ascii.charCodeAt(1); //bez STX
    for(let i=2; i<ascii.length; i++)
        bcc = bcc^ascii.charCodeAt(i);
    return bcc;
}

function encodeID(str){
	var iD = parseInt(str);
	//TODO: group addressing
	if(iD)
		iD = iD + 0x40;
	else iD = 0x2A;
    return String.fromCharCode(iD);
}

function decodeID(ascii){
	var decoded = ascii.charCodeAt(0);
	if(decoded >= 0x41)
		decoded = decoded - 0x40;
	else //group ID
		decoded = decoded + 0x10;
	return decoded;
}

function decimal2Hex(d, width) {
	var hex = Number(d).toString(16);
	if(!width) width = 2;
	while (hex.length < width) {
        hex = "0" + hex;
    }
    return hex.toUpperCase();
}

function asciiHex2str(str){
	let ascii = '';
	for(let l=0; l<str.length; l=l+2){
		let i = parseInt(str[l] + str[l+1], 16);
		let ch = String.fromCharCode(i);
		if(i>0)
			ascii += ch;
	}
	return ascii;
}


module.exports = NECD;